@@ -751,7 +751,7 @@ struct XLA_FFI_Api {
   XLA_FFI_Extension_Base* extension_start;
 
   XLA_FFI_Api_Version api_version;
-  XLA_FFI_InternalApi* internal_api;
+  const XLA_FFI_InternalApi* internal_api;
 
   _XLA_FFI_API_STRUCT_FIELD(XLA_FFI_Error_Create);
   _XLA_FFI_API_STRUCT_FIELD(XLA_FFI_Error_GetMessage);
